 /*
 * xircom.c rev 0.12b 01/08/2001
 * BeOS Ethernet Driver for the 16-Bit Xircom PCMCIA-cards
 *
 * Cards tested and supported:
 *   -Xircom Realport RE-10 (CE3 type)
 * Cards that might work, but are not tested:
 *   -Xircom CE2, CEM28, CEM33, CE33, CE3, CEM56, REM56 cards (Ethernet only)
 * Cards that won't work:
 *   -Old CE Cards (Support NOT planned)
 *   -All Xircom Cardbus cards (Support planned, Development started)
 *   -"OEM-Xircom" cards of manufacturers like Intel and Compaq (Support planned)
 * 
 * Written and maintained by Martijn Coenen <M.F.A.Coenen@student.tue.nl>
 * Copyright (c) 2000,2001 Martijn Coenen
 * 
 * THANKS TO:
 * - The developers of the linux driver (this is merely a port)
 * - hank@be.com  : Sample code, review of the code, cardbus code
 * - Peter King   : Sent me quite some hardware to work and test with
 * - Dean Ritchie : Work on Cardbus code
 * - Everyone sending feedback about the driver
 *
 * Based on tcm509/ethertemplate driver
 * CopyRight (c) 2000 Be Inc., All Rights Reserved. 
 * 
 * Most parts of the hardware specific code is taken straight from the linux xirc2ps
 * driver. There is often simply no reason to change it.
 * I've been warned about eventual licensing-problems by using the linux driver, but
 * according to the ALTERNATIVE license of the linux driver (which you may read below),
 * I am playing the game by the rules. I wrote the current maintainer of the
 * linux driver, to be sure about any licensing problems, but did not get a
 * reply. 
 * If there IS a problem, I wonder why the current maintainer didn't answer
 * my question...(or why my mailer didn't work) :)
 *
 * I will print the entire copyright notice of the linux driver between the dash-lines.
 *
 * ------------------------------------------------------------------------
 * ------------------------------------------------------------------------
 * 
 * [xirc2ps_cs.c wk 03.11.99] (1.40 1999/11/18 00:06:03)
 * Xircom CreditCard Ethernet Adapter IIps driver
 * Xircom Realport 10/100 (RE-100) driver 
 *
 * This driver originally was made by Werner Koch. Since the driver was left
 * unmaintained for some time, there have been some improvements and changes
 * since. These include supporting some of the "Realport" cards and develop-
 * ing enhancements to support the new ones.
 * It is made for CE2, CEM28, CEM33, CE33 and 
 * CEM56 cards. The CEM56 cards work both with their modem and ethernet
 * interface. The RealPort 10/100 Modem and similar cards are supported but
 * with some bugs which are being corrected as they are detected. 
 * 
 * Code revised and maintained by Allan Baker Ortegon
 * al527261@prodigy.net.mx
 * Written originally by Werner Koch based on David Hinds' skeleton of the
 * PCMCIA driver. The code has been modified as to make the newer cards
 * available.
 *
 * The latest code for the driver, information on the development project
 * for the Xircom RealPort and CE cards for the PCMCIA driver, and other
 * related material, can be found at the following URL, which is underway:
 * 
 * "http://xirc2ps.linuxbox.com/index.html"
 *
 * Any bugs regarding this driver, please send them to:
 * alanyuu@linuxbox.com
 *
 * The driver is still evolving and there are many cards which will benefit
 * from having alpha testers. If you have a particular card and would like
 * to be involved in this ongoing effort, please send mail to the maintainer.
 * 
 * Special thanks to David Hinds, to Xircom for the specifications and their
 * software development kit, and all others who may have colaborated in the
 * development of the driver: Koen Van Herck (Koen.Van.Herck@xircom.com),
 * 4PC GmbH Duesseldorf, David Luger, et al.
 * 
 *
 ************************************************************************
 * Copyright (c) 1997,1998 Werner Koch (dd9jn)
 * Copyright (c) 1999 Allan Baker Ortegon 
 *
 * This driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * ALTERNATIVELY, this driver may be distributed under the terms of
 * the following license, in which case the provisions of this license
 * are required INSTEAD OF the GNU General Public License.  (This clause
 * is necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ------------------------------------------------------------------------
 * ------------------------------------------------------------------------
 *
 * The BeOS driver may be distributed under the same terms as the ALTERNATIVE
 * (NOT the GPL) license mentioned above, and carries the same disclaimer:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */ 

#include <OS.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <module.h>
#include <malloc.h>
#include <SupportDefs.h>
#include <ByteOrder.h>
#include "ether_driver.h"
#include <driver_settings.h>

/* Include PCMCIA headers */

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>

#define MAX_DEVS        8
static dev_link_t *     devs[MAX_DEVS];
static int              ndevs = 0;
#define DEVNAME_LENGTH		64			
#define MAX_CARDS 			 4
#define ETH_ZLEN			60
 
#define kDevName "xircom"

#define kDevDir "net/" kDevName "/"

#define DEVNAME_LENGTH		64			

/* debug flags */
#define ERR       0x0001
#define INFO      0x0002
#define RX        0x0004		/* dump received frames */
#define TX        0x0008		/* dump transmitted frames */
#define INTERRUPT 0x0010		/* interrupt calls */
#define FUNCTION  0x0020		/* function calls */
#define BUS_IO    0x0040		/* reads and writes across the [pci,isa,pcmcia,cardbus] bus */
#define SEQ		  0x0080		/* trasnmit & receive TCP/IP sequence sequence numbers */
#define WARN	  0x0100		/* Warnings - off for final release */

/* diagnostic debug flags - compile in here or set while running with debugger "xircom" command */
#define DEBUG_FLAGS ( ERR | INFO  )

#define BUFFER_SIZE			2048L
#define MAX_FRAME_SIZE		1514			/* 1514 + 4 bytes checksum */

static u_long irq_mask = 0xdeb8;
static char 				*gDevNameList[MAX_CARDS+1];
static int32 				gOpenMask = 0;

struct ether_dev_info  {
	uint32	irq;
	uint32 	port;
} gDevInfo[MAX_CARDS+1];

typedef struct ether_dev_info ether_dev_info;

#define output_wait(data, t)    acquire_sem_etc(data->olock, 1, B_TIMEOUT, t)
#define output_unwait(data, c)	release_sem_etc(data->olock, c, B_DO_NOT_RESCHEDULE)

#define write8(io_addr, byte) 	(*pci->write_io_8)((io_addr), (byte))
#define write16(io_addr, word) 	(*pci->write_io_16)((io_addr), (word))
#define write32(io_addr, lword) (*pci->write_io_32)((io_addr), (lword))
#define read8(io_addr) 			(*pci->read_io_8)(io_addr)
#define read16(io_addr) 		(*pci->read_io_16)(io_addr)
#define read32(io_addr) 		(*pci->read_io_32)(io_addr)

/* Driver Entry Points */
__declspec(dllexport) status_t init_hardware(void );
__declspec(dllexport) status_t init_driver(void );
__declspec(dllexport) void uninit_driver(void );
__declspec(dllexport) const char** publish_devices(void );
__declspec(dllexport) device_hooks *find_device(const char *name );


/* Hardware specific definitions */

#define XIRCREG_CR  0	/* Command register (wr) */
enum xirc_cr {
    TransmitPacket = 0x01,
    SoftReset = 0x02,
    EnableIntr = 0x04,
    ForceIntr  = 0x08,
    ClearTxFIFO = 0x10,
    ClearRxOvrun = 0x20,
    RestartTx	 = 0x40
};
#define XIRCREG_ESR 0	/* Ethernet status register (rd) */
enum xirc_esr {
    FullPktRcvd = 0x01, /* full packet in receive buffer */
    PktRejected = 0x04, /* a packet has been rejected */
    TxPktPend = 0x08,	/* TX Packet Pending */
    IncorPolarity = 0x10,
    MediaSelect = 0x20	/* set if TP, clear if AUI */
};
#define XIRCREG_PR  1	/* Page Register select */
#define XIRCREG_EDP 4	/* Ethernet Data Port Register */
#define XIRCREG_ISR 6	/* Ethernet Interrupt Status Register */
enum xirc_isr {
    TxBufOvr = 0x01,	/* TX Buffer Overflow */
    PktTxed  = 0x02,	/* Packet Transmitted */
    MACIntr  = 0x04,	/* MAC Interrupt occured */
    TxResGrant = 0x08,	/* Tx Reservation Granted */
    RxFullPkt = 0x20,	/* Rx Full Packet */
    RxPktRej  = 0x40,	/* Rx Packet Rejected */
    ForcedIntr= 0x80	/* Forced Interrupt */
};
#define XIRCREG1_IMR0 12 /* Ethernet Interrupt Mask Register (on page 1)*/
#define XIRCREG1_IMR1 13
#define XIRCREG0_TSO  8  /* Transmit Space Open Register (on page 0)*/
#define XIRCREG0_TRS  10 /* Transmit reservation Size Register (page 0)*/
#define XIRCREG0_DO   12 /* Data Offset Register (page 0) (wr) */
#define XIRCREG0_RSR  12 /* Receive Status Register (page 0) (rd) */
enum xirc_rsr {
    PhyPkt = 0x01,	/* set:physical packet, clear: multicast packet */
    BrdcstPkt = 0x02,	/* set if it is a broadcast packet */
    PktTooLong = 0x04,	/* set if packet length > 1518 */
    AlignErr = 0x10,	/* incorrect CRC and last octet not complete */
    CRCErr = 0x20,	/* incorrect CRC and last octet is complete */
    PktRxOk = 0x80	/* received ok */
};
#define XIRCREG0_PTR 13 /* packets transmitted register (rd) */
#define XIRCREG0_RBC 14 /* receive byte count regsister (rd) */
#define XIRCREG1_ECR 14 /* ethernet configurationn register */
enum xirc_ecr {
    FullDuplex = 0x04,	/* enable full duplex mode */
    LongTPMode = 0x08,	/* adjust for longer lengths of TP cable */
    DisablePolCor = 0x10,/* disable auto polarity correction */
    DisableLinkPulse = 0x20, /* disable link pulse generation */
    DisableAutoTx = 0x40, /* disable auto-transmit */
};
#define XIRCREG2_RBS 8	/* receive buffer start register */
#define XIRCREG2_LED 10 /* LED Configuration register */
/* values for the leds:    Bits 2-0 for led 1
 *  0 disabled		   Bits 5-3 for led 2
 *  1 collision
 *  2 noncollision
 *  3 link_detected
 *  4 incor_polarity
 *  5 jabber
 *  6 auto_assertion
 *  7 rx_tx_activity
 */
#define XIRCREG2_MSR 12 /* Mohawk specific register */

#define XIRCREG4_GPR0 8 /* General Purpose Register 0 */
#define XIRCREG4_GPR1 9 /* General Purpose Register 1 */
#define XIRCREG2_GPR2 13 /* General Purpose Register 2 (page2!)*/
#define XIRCREG4_BOV 10 /* Bonding Version Register */
#define XIRCREG4_LMA 12 /* Local Memory Address Register */
#define XIRCREG4_LMD 14 /* Local Memory Data Port */
/* MAC register can only by accessed with 8 bit operations */
#define XIRCREG40_CMD0 8    /* Command Register (wr) */
enum xirc_cmd { 	    /* Commands */
    Transmit = 0x01,
    EnableRecv = 0x04,
    DisableRecv = 0x08,
    Abort = 0x10,
    Online = 0x20,
    IntrAck = 0x40,
    Offline = 0x80
};
#define XIRCREG5_RHSA0	10  /* Rx Host Start Address */
#define XIRCREG40_RXST0 9   /* Receive Status Register */
#define XIRCREG40_TXST0 11  /* Transmit Status Register 0 */
#define XIRCREG40_TXST1 12  /* Transmit Status Register 10 */
#define XIRCREG40_RMASK0 13  /* Receive Mask Register */
#define XIRCREG40_TMASK0 14  /* Transmit Mask Register 0 */
#define XIRCREG40_TMASK1 15  /* Transmit Mask Register 0 */
#define XIRCREG42_SWC0	8   /* Software Configuration 0 */
#define XIRCREG42_SWC1	9   /* Software Configuration 1 */
#define XIRCREG42_BOC	10  /* Back-Off Configuration */
#define XIRCREG44_TDR0	8   /* Time Domain Reflectometry 0 */
#define XIRCREG44_TDR1	9   /* Time Domain Reflectometry 1 */
#define XIRCREG44_RXBC_LO 10 /* Rx Byte Count 0 (rd) */
#define XIRCREG44_RXBC_HI 11 /* Rx Byte Count 1 (rd) */
#define XIRCREG45_REV	 15 /* Revision Register (rd) */
#define XIRCREG50_IA	8   /* Individual Address (8-13) */

#ifndef MANFID_XIRCOM
  #define MANFID_XIRCOM 	   0x0105
#endif

/* card types */
#define XIR_UNKNOWN  0	/* unknown: not supported */
#define XIR_CE	     1	/* (prodid 1) different hardware: not supported */
#define XIR_CE2      2	/* (prodid 2) */
#define XIR_CE3      3	/* (prodid 3) */
#define XIR_CEM      4	/* (prodid 1) different hardware: not supported */
#define XIR_CEM2     5	/* (prodid 2) */
#define XIR_CEM3     6	/* (prodid 3) */
#define XIR_CEM33    7	/* (prodid 4) */
#define XIR_CEM56M   8	/* (prodid 5) */
#define XIR_CEM56    9	/* (prodid 6) */
#define XIR_CM28    10	/* (prodid 3) modem only: not supported here */
#define XIR_CM33    11	/* (prodid 4) modem only: not supported here */
#define XIR_CM56    12	/* (prodid 5) modem only: not supported here */
#define XIR_CG	    13	/* (prodid 1) GSM modem only: not supported */
#define XIR_CBE     14	/* (prodid 1) cardbus ethernet: not supported */
/*====================================================================*/


/* This assumes that in every func we use SelectPage ioaddr is available!*/
#define SelectPage(pgnr)   write8(ioaddr + XIRCREG_PR,(pgnr))


#define RX_BUFFERS  32			/* must be a power of 2 */
#define RX_BUFFER_MASK  (RX_BUFFERS-1)
	

/* per driver intance globals */

typedef struct local_info_t {
    dev_node_t	node;
    uint8		macAddr[6];	/* Mac addr only available via CIS?*/
    int			stop;
	int			card_type; /* Store the kind of card here */
	int			mohawk; /* If it's a mohawk type (CE3) */
	int			modem; /* If a card has a modem */
	int			dingo; /* If it's a CEM56 type */
	caddr_t		dingo_ccr;
	int			silrev; /* Silicon revision */
	int			new_mii; /* To indicate whether there's an mii */
	int			probe_port; /* Whether autoneg should be used or not */
} local_info_t;


typedef struct {
	uint8			macAddr[6];								/* ethernet mac address */
	uint16			irq;									/* our IRQ line */
	int32			devID; 									/* device identifier: 0-n */
	sem_id			ilock, olock;							/* i/o semaphores */
	int32			readLock, writeLock;					/* reentrant reads/writes not allowed */
	int32			blockFlag;								/* for blocking or nonblocking reads */
	uint32			port; 									/* Base address of hostRegs */
	uint16			media;									/* speed duplex media settings */
	uint16			speed;
	uint16			duplex;
	uint32			debug;									/* debuging level */
	area_id			rx_buf_area; 							/* receive descriptor and buffer areas */
	uchar			*rx_buf[RX_BUFFERS];					/* receive buffers */
	uint32			rx_desc[RX_BUFFERS];					/* rececive buffer descriptors */
	int16			rx_received, rx_acked;					/* index to in use tx & rx ring buffers */
	uint32			MultiOn;								/* receive multicast frames */
	uint32			PromiscOn;								/* receive multicast frames */
	local_info_t	*local;									
	/* statistics */
	uint32			handled_int, unhandled_int;				/* interrupts rececived & handled */
	uint32			stats_rx, stats_tx;						/* receive & transmit frames counts */
	uint32			stats_rx_overrun, stats_tx_underrun;
	uint32			stats_rx_len_err, stats_tx_len_err;
	uint32			stats_rx_frame_err, stats_rx_crc_err;
	uint32			stats_collisions;  						/* error stats */

} ether_dev_info_t;


//#define ETHER_DEBUG(flags, run_time_flags, args...) if (flags & run_time_flags) dprintf(args)
void ETHER_DEBUG(ether_dev_info_t * device, int32 flags, char * format, ...) {
		if (device->debug & flags) {
		va_list		args;
		char		s[4096];
		va_start(args, format);
		vsprintf( s, format, args );
		va_end(args);
		dprintf("%s",s);
	}
}

/* for serial debug command*/
#define DEBUGGER_COMMAND true
#if DEBUGGER_COMMAND
ether_dev_info_t * gdev;
static int 		xircom(int argc, char **argv);
#endif


/* prototypes */
static status_t open_hook(const char *name, uint32 flags, void **_cookie);
static status_t close_hook(void *_device);
static status_t free_hook(void *_device);
static status_t control_hook(void * cookie,uint32 msg,void *buf,size_t len);
static status_t read_hook(void *_device,off_t pos, void *buf,size_t *len);
static status_t write_hook(void *_device, off_t pos, const void *buf, size_t *len);
static status_t writev_hook( void *_device, off_t position, const iovec *vec, size_t count, size_t *length );
static int32    interrupt_hook(void *_device);                	   /* interrupt handler */

static void     get_mac_address(ether_dev_info_t *device);               /* get ethernet address */
static status_t setpromisc(ether_dev_info_t * device, uint32 On);        /* set hardware to receive all packets */
static status_t domulti(ether_dev_info_t *device, uint8 *addr);		   /* add multicast address to hardware filter list */
static void     reset_device(ether_dev_info_t *device);                  /* reset the lan controller (NIC) hardware */
static void 	dump_packet(const char * msg, unsigned char * buf, uint16 size); /* diagnostic packet trace */
static void 	dump_rx_desc(ether_dev_info_t *device);				   /* dump hardware receive descriptor structures */
static void 	dump_tx_desc(ether_dev_info_t *device);				   /* dump hardware transmit descriptor structures */
static void 	dump_window(ether_dev_info_t *device, uchar window);	   /* dump hardware registers */
static void		clear_statistics(ether_dev_info_t *device);			   /* clear statistics counters */
static status_t allocate_resources(ether_dev_info_t *device);     	   /* allocate semaphores & spinlocks */
static void     free_resources(ether_dev_info_t *device);                /* deallocate semaphores & spinlocks */
static status_t init_ring_buffers(ether_dev_info_t *device);
static void		free_ring_buffers(ether_dev_info_t *device);
static uint8 	load_irq_settings(uint16 *irq, uint32 *port);

/* MII management prototypes */

static void		mii_idle(ioaddr_t ioaddr);
static void 	mii_putbit(ioaddr_t ioaddr, unsigned data);
static int  	mii_getbit(ioaddr_t ioaddr);
static void 	mii_wbits(ioaddr_t ioaddr, unsigned data, int len);
static unsigned mii_rd(ioaddr_t ioaddr,	u_char phyaddr, u_char phyreg);
static void 	mii_wr(ioaddr_t ioaddr, u_char phyaddr, u_char phyreg, unsigned data, int len);
static int 		init_mii(ether_dev_info_t *device);

status_t pc_init_driver(void);
void pc_uninit_driver(void);
const char **pc_publish_devices(void);

static device_hooks gDeviceHooks =  {
	open_hook, 			/* -> open entry point */
	close_hook,         /* -> close entry point */
	free_hook,          /* -> free entry point */
	control_hook,       /* -> control entry point */
	read_hook,          /* -> read entry point */
	write_hook,         /* -> write entry point */
	NULL,               /* -> select entry point */
	NULL,               /* -> deselect entry point */
	NULL,               /* -> readv */
	writev_hook         /* -> writev */
};


/* ------------------------------------------------------------------ *
 * Driver Entry Points
/* ------------------------------------------------------------------ */

status_t init_hardware (void)
{
	return B_NO_ERROR;
}


char				*pci_name = B_PCI_MODULE_NAME;
pci_module_info		*pci;


status_t init_driver()
{
	status_t 		status;
	int32			entries;
	char			devName[64];
	int32			i;
	
	status = get_module(pci_name, (module_info **)&pci);
	if(status != B_NO_ERROR)
		goto err1;

	if ((status = pc_init_driver()) != B_OK) {
		goto err2;
	};

	return B_NO_ERROR;
	
err2:
	pc_uninit_driver();
	put_module(pci_name);
err1:
	printk("pc_init_card failed");
	return status;
}

void uninit_driver(void)
{
	int32 	i;
	void 	*item;

	pc_uninit_driver();

}


device_hooks *find_device(const char *name)
{
	int32 	i;
	char 	*item;

// to do: fix for multiple pc cards

	return &gDeviceHooks;

}


static status_t open_hook(const char *name, uint32 flags, void **cookie) {

	int32				devID;
	int32				mask;
	status_t			status;
	local_info_t		*tempdev;	
	char 				*devName;
	ether_dev_info_t 	*device;
	uint16 				i;
	unsigned			value;
	int 				nd;
	ioaddr_t 			ioaddr;
	
	devID=0; // to do: fix for multiple instances
	/* Check if the device is busy and set in-use flag if not */
	mask = 1 << devID;
	if (atomic_or(&gOpenMask, mask) &mask) {
		dprintf(kDevName " busy\n");
		return B_BUSY;
	}

	/*	Allocate storage for the cookie*/
	if (!(*cookie = device = (ether_dev_info_t *)malloc(sizeof(ether_dev_info_t)))) {
		status = B_NO_MEMORY;
		goto err0;
	}
	memset(device, 0, sizeof(ether_dev_info_t));
	
    for (nd = 0; nd < MAX_DEVS; nd++) {
// dprintf("comparing %s to %s\n", devs[nd]->dev->dev_name, name);
   if ((devs[nd] != NULL) &&
        (strcmp(devs[nd]->dev->dev_name, name) == 0))
        break;
    }
	if (nd != MAX_DEVS){
		device->irq = devs[nd]->irq.AssignedIRQ;
		device->port = devs[nd]->io.BasePort1;	
		device->local = devs[nd]->priv;
		ioaddr = devs[nd]->io.BasePort1;
		/* This is ugly...I can't get seem to get MAC-address elsewhere*/
		tempdev = devs[nd]->priv;	
		for (i = 0; i < 6; i++) {
			device->macAddr[i] = tempdev->macAddr[i];
		}
		
		dprintf("xircom: open_hook:: %s: irq %d / iobase 0x%04x", devs[nd]->dev->dev_name, device->irq, device->port);	
		dprintf("\n");
	} else {
		dprintf("xircom:open_hook:: err dev not found \n");
   		free(device);
		status=B_ERROR;
		goto err1;
	}
		
	device->debug = DEBUG_FLAGS;


	#if DEBUGGER_COMMAND
	gdev = device;
	add_debugger_command (kDevName, xircom, "Ethernet driver Info");
	#endif
	
	ETHER_DEBUG(device, FUNCTION, kDevName ": open %s device=%x ", name, device);

	/* Allocate resources */
	if (allocate_resources(device) != B_OK) {
		goto err1;
	}	
	/* Reset the hardware */

	reset_device(device);	

	/* Initialize the ring buffers */
	if (init_ring_buffers(device) != B_OK)
		goto err1;
	
	
	/* Install interrupt handler */
	install_io_interrupt_handler(device->irq, interrupt_hook, *cookie, 0 ); 

 	return B_NO_ERROR;
		
	err1:
		free_resources(device);
		free(device);	
	
	err0:
		atomic_and(&gOpenMask, ~mask);
		ETHER_DEBUG(device, ERR,  " open failed!\n");
		return status;
}

static status_t close_hook(void *_device) {
	ether_dev_info_t *device = (ether_dev_info_t *) _device;
	ioaddr_t ioaddr = device->port;

	ETHER_DEBUG(device, FUNCTION, kDevName ": close_hook\n");

	SelectPage(0);
    write8(ioaddr + XIRCREG_CR, 0);  /* disable interrupts */
    SelectPage(0x01);
    write8(ioaddr + XIRCREG1_IMR0, 0x00); /* forbid all ints */
    SelectPage(4);
    write8(ioaddr + XIRCREG4_GPR1, 0);	/* clear bit 0: power down */
    SelectPage(0);

#if DEBUGGER_COMMAND
	remove_debugger_command (kDevName, xircom);
#endif
	
	free_resources(device);

	return (B_NO_ERROR);
}


static status_t free_hook(void * cookie) {
	ether_dev_info_t *device = (ether_dev_info_t *) cookie;


	ETHER_DEBUG(device, FUNCTION, kDevName": free %x\n",device);
    /* Remove Interrupt Handler */
	remove_io_interrupt_handler( device->irq, interrupt_hook, cookie );

	free_ring_buffers(device);
	
	/* Device is now available again */
	atomic_and(&gOpenMask, ~(1 << device->devID));

	free(device);
	return 0;
}


/* Standard driver control function */
static status_t control_hook(void * cookie,uint32 msg,void *buf,size_t len)
{
	ether_dev_info_t *device = (ether_dev_info_t *) cookie;

	switch (msg) {
		case ETHER_GETADDR: {
			uint8 i;
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x ether_getaddr\n",msg);
			for (i=0; i<6; i++) {
				((uint8 *)buf)[i] = device->macAddr[i];
			}
			return B_NO_ERROR;
		}
		case ETHER_INIT:
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x init\n",msg);
			return B_NO_ERROR;
			
		case ETHER_GETFRAMESIZE:
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x get_framesize\n",msg);
			*(uint32 *)buf = MAX_FRAME_SIZE;
			return B_NO_ERROR;
			
		case ETHER_ADDMULTI:
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x add multi\n",msg);
			return (domulti(device, (unsigned char *)buf));
		
		case ETHER_SETPROMISC:
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x set promiscuous\n",msg);
			return setpromisc(device, *(uint32 *)buf);
		
		case ETHER_NONBLOCK:
			ETHER_DEBUG(device, FUNCTION, kDevName ": control %x blocking\n",msg);

			if (*((int32 *)buf))
				device->blockFlag = B_TIMEOUT;
			else
				device->blockFlag = 0;
			return B_NO_ERROR;
			
		default:
			ETHER_DEBUG(device, ERR, kDevName ": unknown iocontrol %x\n",msg);
			return B_ERROR;
	}

}


/* The read() system call - upper layer interface to the ethernet driver */
static status_t  read_hook(void *_device,off_t pos, void *buf,size_t *len)
{
	ether_dev_info_t  *device = (ether_dev_info_t *) _device;
	status_t    status;
	int32		count = 0;
	
		
	/* Block until data is available */
	if ((status = acquire_sem_etc(device->ilock, 1, B_CAN_INTERRUPT | device->blockFlag, 0)) != B_NO_ERROR) {
		*len = 0;
		return status;
	}
	
	/* Prevent reentrant read */
	if (atomic_or(&device->readLock, 1 )) {
	    release_sem_etc( device->ilock, 1, 0 );
		*len = 0;
		return B_ERROR;
	}
	/* Copy data from device to buf */
	*len = device->rx_desc[device->rx_acked];
	
	memcpy(buf, device->rx_buf[device->rx_acked], *len);
	
	device->rx_acked = (device->rx_acked + 1) & RX_BUFFER_MASK;
	
	if (device->debug & RX) {
		if (*len)
			dump_packet("rx", buf, *len);
	}
	device->readLock = 0; /* release reantrant lock */
	return B_OK;
}


/*
 * The write() system call - upper layer interface to the ethernet driver
 */
 
static status_t write_hook(void *_device, off_t pos, const void *buf, size_t *len)
{
	ether_dev_info_t  *device = (ether_dev_info_t *) _device;
	int16       frame_size;
	status_t    status;
	int 		okay;
    unsigned 	freespace;
    unsigned 	pktlen = *len;
	int16       tx_needed;

	ioaddr_t	ioaddr = device->port;
	ETHER_DEBUG(device, FUNCTION, kDevName ":write buf %x len %d (decimal)\n",buf,*len);
	 
	if( *len > MAX_FRAME_SIZE ) {
		ETHER_DEBUG(device, ERR, kDevName ": write %d > MAX_FRAME_SIZE tooo long\n",*len);
		*len = MAX_FRAME_SIZE;
	}

	/* Prevent reentrant write */
	if (atomic_or( &device->writeLock, 1 )) {
//		ETHER_DEBUG(device, ERR, "write_hook: reentrant write\n"); /* this should never happen */
		snooze(5000);
		*len = 0;
		return B_ERROR;
	}
    
	/* Try and acquire TX sem */
	if( (status = acquire_sem_etc( device->olock, 1, B_CAN_INTERRUPT, 0 )) != B_NO_ERROR ) {
		*len = 0;
		device->writeLock = 0;
		return B_ERROR;
	}	

    /* Set minimum required len */
	if (pktlen < ETH_ZLEN)
		pktlen = ETH_ZLEN;


    SelectPage(0);
    /* Next necessary ? */
    write16(ioaddr + XIRCREG0_TRS, (u_short)pktlen+2);
    freespace = read16(ioaddr + XIRCREG0_TSO);
    okay = freespace & 0x8000;
    freespace &= 0x7fff;
    okay = pktlen + 2 < freespace;
    if (!okay) { /* not enough space */
		*len = 0;
		device->writeLock = 0;
		return B_ERROR;				
	}
	write16(ioaddr + XIRCREG_EDP, (u_short)pktlen);
	/* send the packet, want this 16-bit for mohawk-only because of odd packet problems*/
	if (device->local->mohawk) {
    		int i;
    		uint16 *bufL = (uint16 *) buf;
    		for (i=0;i< (pktlen +1) >> 1;i++)
				write16(ioaddr + XIRCREG_EDP, *bufL++);
	}
	else { /* No mohawk, stick to 8-bit writes FIXME! */
			int i;
    		uint8 *bufL = (uint8 *) buf;
    		for (i=0;i<pktlen;i++)
				write8(ioaddr + XIRCREG_EDP, *bufL++);
		
	}
	/* Re-enable interrupts for mohawk*/
	
	if (device->local->mohawk) {
		write8(ioaddr + XIRCREG_CR, TransmitPacket|EnableIntr); 
	}
	/* Another write may now take place */
    device->writeLock = 0;
	return B_OK;
}
	
static status_t writev_hook( void *_device, off_t position, const iovec *vec, size_t count, size_t *length ) {

	return B_ERROR;
};


/* service interrupts generated by the Lan controller (card) hardware */
static int32
interrupt_hook(void *_device)
{
	ether_dev_info_t *device = (ether_dev_info_t *) _device;
	int32 handled = B_UNHANDLED_INTERRUPT;
	int32 status;
    int32 count;
	ioaddr_t ioaddr = device->port;
	local_info_t *local = device->local;
	u_char saved_page;
	unsigned bytes_rcvd;
    unsigned int_status, eth_status, rx_status, tx_status;
    unsigned rsr, pktlen;

	ETHER_DEBUG(device, INTERRUPT, kDevName ": ISR %x\n", status);	
	handled = B_INVOKE_SCHEDULER;
	device->handled_int++;  						/* update stats */
    /* Disable interrupt, only for mohawk*/
	if (device->local->mohawk)
		write8(ioaddr + XIRCREG_CR, 0);

	saved_page = read8(ioaddr + XIRCREG_PR);
    /* Read the ISR to see whats the cause for the interrupt.
     * This also clears the interrupt flags on CE2 cards
     */
    int_status = read8(ioaddr + XIRCREG_ISR);
	eth_status = read8(ioaddr + XIRCREG_ESR);

    SelectPage(0x40);
    rx_status  = read8(ioaddr + XIRCREG40_RXST0);
    write8(ioaddr + XIRCREG40_RXST0, (~rx_status & 0xff));
    tx_status = read8(ioaddr + XIRCREG40_TXST0);
    tx_status |= read8(ioaddr + XIRCREG40_TXST1) << 8;
    write8(ioaddr + XIRCREG40_TXST0, 0);
    write8(ioaddr + XIRCREG40_TXST1, 0);

	/*Read interrupt*/
	SelectPage(0);
    /* FIXME: We only read one packet in an interrupt!*/
    if (eth_status & FullPktRcvd) {
    rsr = read8(ioaddr + XIRCREG0_RSR);
	if (rsr & PktRxOk) {
		pktlen = read16(ioaddr + XIRCREG0_RBC);
	    bytes_rcvd += pktlen;
	    /* Get the packet */
		if (local->silrev == 0) {
			unsigned rhsa; /* receive start address */

			SelectPage(5);
		    rhsa = read16(ioaddr + XIRCREG5_RHSA0);
		    SelectPage(0);
		    rhsa += 3; /* skip control infos */
		
		    if (rhsa >= 0x8000)
				rhsa = 0;
		
		    if (rhsa + pktlen > 0x8000) {
				unsigned i;
				uint8 *rx_buf = (uint8 *) device->rx_buf[device->rx_received];
				if (pktlen > MAX_FRAME_SIZE)
					pktlen = min(MAX_FRAME_SIZE, pktlen);
		
				device->rx_desc[device->rx_received] = pktlen;
		
				for (i=0; i < pktlen ; i++, rhsa++) {
			    	rx_buf[i] = read8(ioaddr + XIRCREG_EDP);
			    	if (rhsa == 0x8000) {
						rhsa = 0;
						i--;
				   	}
		    	}
		    }
		    else {
				unsigned i;
				uint16 *rx_buf = (uint16 *) device->rx_buf[device->rx_received];
				if (pktlen > MAX_FRAME_SIZE)
					pktlen = min(MAX_FRAME_SIZE, pktlen);
		
				device->rx_desc[device->rx_received] = pktlen;
			
				for (i=0; i < (pktlen + 1) >> 1; i++)
					*rx_buf++ = read16(ioaddr + XIRCREG_EDP);
			}
		}
		
		else { /* This includes mohawk */
			unsigned i;
			uint16 *rx_buf = (uint16 *) device->rx_buf[device->rx_received];
			if (pktlen > MAX_FRAME_SIZE)
				pktlen = min(MAX_FRAME_SIZE, pktlen);
		
			device->rx_desc[device->rx_received] = pktlen;
			for (i=0; i < (pktlen + 1) >> 1; i++)
				*rx_buf++ = read16(ioaddr + XIRCREG_EDP);
			
		}
		device->rx_received = (device->rx_received + 1) & RX_BUFFER_MASK;
		release_sem_etc( device->ilock, 1, B_DO_NOT_RESCHEDULE );
		device->stats_rx++;	
	
		/* Move to next packet */				
		write32(ioaddr + XIRCREG0_DO, 0x8000); /* issue cmd: skip_rx_packet */
	
	} else {
	dprintf("Packet receive error\n");
	 /* Packet not ok */
	}
	if (rsr & PktTooLong) {
	dprintf("Packet receive error\n");
	}
	if (rsr & CRCErr) {
	dprintf("Packet receive error\n");
	}
	if (rsr & AlignErr) {
	dprintf("Packet receive error\n");
	}
	/* Read in new eth_status for next loop */
	eth_status = read8(ioaddr + XIRCREG_ESR);
}

    if (rx_status & 0x10) { 
    	/* RX overrun */
        dprintf("Packet receive error\n");
	
		write8(ioaddr + XIRCREG_CR, ClearRxOvrun);
	}
		
	
	/*Write interrupt*/
	
	if (int_status & PktTxed) {
		/* Packet succesfully transmitted */
		release_sem_etc( device->olock, 1, B_DO_NOT_RESCHEDULE );		
		get_sem_count(device->olock, &count);
		
		}
    if (tx_status & 0x0002) {	/* Too much collissions */
		dprintf("Too much collissions, Restarting TX");
		write8(ioaddr + XIRCREG_CR, RestartTx);  /* RestartTx */
    }
    if (tx_status & 0x0040) {
		//dprintf("Send error");
	}
    
	SelectPage(saved_page); /* Reset saved-page */
	write8(ioaddr + XIRCREG_CR, EnableIntr);  /* Enable interrupts */
	return handled;
}

static uchar
log2(uint32 mask)
{
  uchar value;

  mask >>= 1;
  for (value = 0; mask; value++)
        mask >>= 1;

  return value;
}

#define RNDUP(x, y) (((x) + (y) - 1) & ~((y) - 1))

static status_t init_ring_buffers(ether_dev_info_t *device)
{
	uint32 	size;
	physical_entry		entry;
	uint16 i;

	/* create rx buffer area */
	size = RNDUP(BUFFER_SIZE * RX_BUFFERS, B_PAGE_SIZE);
	if ((device->rx_buf_area = create_area( kDevName " rx buffers", (void **) device->rx_buf,
		B_ANY_KERNEL_ADDRESS, size, B_FULL_LOCK, B_READ_AREA | B_WRITE_AREA )) < 0) {
		ETHER_DEBUG(device, ERR, kDevName " create rx buffer area failed %x \n", device->rx_buf_area);
		return device->rx_buf_area;
	}

	/* init rx buffer */
	for ( i = 0; i < RX_BUFFERS; i++) {
		device->rx_buf[i] = device->rx_buf[0] + (i * BUFFER_SIZE);
	}

	/* initialize frame indexes */
	device->rx_received = device->rx_acked = 0;

	return B_OK;
}

static void free_ring_buffers(ether_dev_info_t *device) {

		delete_area(device->rx_buf_area);
}


static void get_mac_address(ether_dev_info_t *device) {

	int 	j;
		
	ETHER_DEBUG(device, INFO, kDevName " mac address ");
	
	/* print mac address */
	for (j=0; j<6; j++) {
		dprintf("%2.2x%c", device->macAddr[j], j==5?' ':':');
	}
	dprintf("\n");
}

/* set hardware so all packets are received. */
static status_t setpromisc(ether_dev_info_t * device, uint32 On) {
	ETHER_DEBUG(device, FUNCTION, kDevName ":setpormisc\n");

	/* Set promisc mode */
		
	return(B_ERROR);
}

static status_t domulti(ether_dev_info_t *device, uint8 *addr) {
	
	ETHER_DEBUG(device, FUNCTION,kDevName ": domulti %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n", 
					addr[0],addr[1],addr[2],addr[3],addr[4],addr[5]);
	/* FIXME: Multicast */
		
	return (B_NO_ERROR);
}

/*============== MII Management functions ===============*/

/****************
 * Turn around for read
 */
static void
mii_idle(ioaddr_t ioaddr)
{
    write8(ioaddr +XIRCREG2_GPR2, 0x04|0); /* drive MDCK low */
    udelay(1);
    write8(ioaddr +XIRCREG2_GPR2, 0x04|1); /* and drive MDCK high */
    udelay(1);
}

/****************
 * Write a bit to MDI/O
 */
static void
mii_putbit(ioaddr_t ioaddr, unsigned data)
{
    if (data) {
	write8(ioaddr + XIRCREG2_GPR2, 0x0c|2|0); /* set MDIO */
	udelay(1);
	write8(ioaddr + XIRCREG2_GPR2, 0x0c|2|1); /* and drive MDCK high */
	udelay(1);
    } else {
	write8(ioaddr + XIRCREG2_GPR2, 0x0c|0|0); /* clear MDIO */
	udelay(1);
	write8(ioaddr + XIRCREG2_GPR2, 0x0c|0|1); /* and drive MDCK high */
	udelay(1);
    }
}


/****************
 * Get a bit from MDI/O
 */
static int
mii_getbit(ioaddr_t ioaddr)
{
    unsigned d;

    write8(ioaddr + XIRCREG2_GPR2, 4|0); /* drive MDCK low */
    snooze(1);
    d = read8(ioaddr + XIRCREG2_GPR2); /* read MDIO */
    write8(ioaddr + XIRCREG2_GPR2, 4|1); /* drive MDCK high again */
    snooze(1);
    return d & 0x20; /* read MDIO */
}


static void
mii_wbits(ioaddr_t ioaddr, unsigned data, int len)
{
    unsigned m = 1 << (len-1);
    for (; m; m >>= 1)
	mii_putbit(ioaddr, data & m);
}


static unsigned
mii_rd(ioaddr_t ioaddr,	u_char phyaddr, u_char phyreg)
{
    int i;
    unsigned data=0, m;

    SelectPage(2);
    for (i=0; i < 32; i++)		/* 32 bit preamble */
	mii_putbit(ioaddr, 1);
    mii_wbits(ioaddr, 0x06, 4); 	/* Start and opcode for read */
    mii_wbits(ioaddr, phyaddr, 5);	/* PHY address to be accessed */
    mii_wbits(ioaddr, phyreg, 5);	/* PHY register to read */
    mii_idle(ioaddr);			/* turn around */
    mii_getbit(ioaddr);

    for (m = 1<<15; m; m >>= 1)
	if (mii_getbit(ioaddr))
	    data |= m;
    mii_idle(ioaddr);
    return data;
}


static void
mii_wr(ioaddr_t ioaddr, u_char phyaddr, u_char phyreg, unsigned data, int len)
{
    int i;

    SelectPage(2);
    for (i=0; i < 32; i++)		/* 32 bit preamble */
	mii_putbit(ioaddr, 1);
    mii_wbits(ioaddr, 0x05, 4); 	/* Start and opcode for write */
    mii_wbits(ioaddr, phyaddr, 5);	/* PHY address to be accessed */
    mii_wbits(ioaddr, phyreg, 5);	/* PHY Register to write */
    mii_putbit(ioaddr, 1);		/* turn around */
    mii_putbit(ioaddr, 0);
    mii_wbits(ioaddr, data, len);	/* And write the data */
    mii_idle(ioaddr);
}

static int
init_mii(ether_dev_info_t *device)
{
    local_info_t *local = device->local;
    ioaddr_t ioaddr = device->port;
    unsigned control, status, linkpartner;
    int i;

    status = mii_rd(ioaddr,  0, 1);
    if ((status & 0xff00) != 0x7800) {
    		return 0; /* No MII */
	}
    local->new_mii = (mii_rd(ioaddr, 0, 2) != 0xffff);
    
    if (local->probe_port)
		control = 0x1000; /* auto neg */
    else if (device->media == 4)
		control = 0x2000; /* no auto neg, 100mbs mode */
    else
		control = 0x0000; /* no auto neg, 10mbs mode */
    
    mii_wr(ioaddr,  0, 0, control, 16);
    snooze(100);
    control = mii_rd(ioaddr, 0, 0);

    if (control & 0x0400) {
		local->probe_port = 0;
	return 0;
    }

    if (local->probe_port) {
	/* according to the DP83840A specs the auto negotiation process
	 * may take up to 3.5 sec, so we use this also for our ML6692
	 * Fixme: Better to use a timer here!
	 */
	 /* THIS IS REALLY TERRIBLE! */
	for (i=0; i < 35; i++) {
	    snooze(HZ/10);	 /* wait 100 msec */
	    status = mii_rd(ioaddr,  0, 1);
	    if ((status & 0x0020) && (status & 0x0004))
		break;
	}

	if (!(status & 0x0020)) {
		/* Auto-neg failed */
	    if (!local->new_mii) {
			control = 0x0000;
			mii_wr(ioaddr,  0, 0, control, 16);
			snooze(100);
		SelectPage(0);
		device->media = (read8(ioaddr + XIRCREG_ESR) & MediaSelect) ? 1 : 2;
	    }
	} else {
	    linkpartner = mii_rd(ioaddr, 0, 5);
	    if (linkpartner & 0x0080) {
			device->media = 4;
	    } else
			device->media = 1;
		}
    }
    return 1;
}

static void reset_device(ether_dev_info_t *device) {
	
	ioaddr_t	ioaddr;
	local_info_t *local = device->local;
	ioaddr = device->port;

	/* First do a hard reset as documented */
		
	SelectPage(4);
	snooze(1); 							/* Wait a msec */
	write8(ioaddr + XIRCREG4_GPR1, 0); 	/* shut down */
    snooze(HZ/25);		     			/* Wait for 40 msec */
    if (local->mohawk)
    	write8(ioaddr + XIRCREG4_GPR1, 1);	 /* power up */
    else
    	write8(ioaddr+ XIRCREG4_GPR1, 1 | 4); /* For other cards: also AIC */
    
    snooze(HZ/50);				     /* Wait for 20 msec */
    
    /* Do a software reset */
    
    write8(ioaddr + XIRCREG_CR, SoftReset); 
    snooze(HZ/50);		     /* Wait for 20 msec */
    write8(ioaddr + XIRCREG_CR, 0);	/* Clear bit */
    snooze(HZ/25);		     /* Wait for 40 msec */
    
	if (local->mohawk) {
		/* set pin GP1 and GP2 to output  (0x0c)
		 * set GP1 to low to power up the ML6692 (0x00)
	 	* set GP2 to high to power up the 10Mhz chip  (0x02)
	 	*/
		write8(ioaddr + XIRCREG4_GPR0, 0x0e);
	}
    snooze(HZ/2);		/* Wait for 500ms */
	    
    local->silrev = local->mohawk ? (read8(ioaddr + XIRCREG4_BOV) & 0x70) >> 4
				   : (read8(ioaddr + XIRCREG4_BOV) & 0x30) >> 4;

	local->probe_port = 1; /* Hard code this to 1 for now */
	/* Is media settable by user (like linux modules params)? */
	if (local->probe_port) {
		if (!local->mohawk) {
	    	SelectPage(4);
	    	write8(ioaddr + XIRCREG4_GPR0, 4);
	    	local->probe_port = 0;
		}
    } else if (device->media == 2) { /* enable 10Base2 */
		SelectPage(0x42);
		write8(ioaddr + XIRCREG42_SWC1, 0xC0);
    } else { /* enable 10BaseT */
		SelectPage(0x42);
		write8(ioaddr + XIRCREG42_SWC1, 0x80);
    }
    snooze(HZ/25);		     /* Wait for 40 msec */
	  
	/* setup the ECR */
    SelectPage(1);
    write8(ioaddr + XIRCREG1_IMR0, 0xff); /* All interrupts*/
    write8(ioaddr + XIRCREG1_IMR1, 1	); /* TxUnderrunDetect */

	SelectPage(0x42);
    write8(ioaddr + XIRCREG42_SWC0, 0x20);

	if (local->silrev != 1) {
		/* Set the border between rx and tx buffer in local memory */
		SelectPage(2);
		write16(ioaddr + XIRCREG2_RBS, 0x2000);
	}
	{	
	/* I don't really mind multicast support just yet - just fill
	 * the cards buffers with it's own ethernet address */
    char *bufaddr;
    int i,j;

    SelectPage(0x50);
    for (i=0,j=8;i<6 ; i++, j++) {
		bufaddr = device->macAddr;
		if (local->mohawk)
			write8(ioaddr + j, bufaddr[5-i]);
		else
			write8(ioaddr + j, bufaddr[i]);			
	}
	SelectPage(0);
	}
	
	SelectPage(0);
    write16(ioaddr + XIRCREG0_DO, 0x2000); /* change offset command, off=0 */

    /* setup MAC IMRs and clear status registers */
    SelectPage(0x40);		     /* Bit 7 ... bit 0 */
    write8(ioaddr + XIRCREG40_RMASK0, 0xff); /* ROK, RAB, rsv, RO, CRC, AE, PTL, MP */
    write8(ioaddr + XIRCREG40_TMASK0, 0xff); /* TOK, TAB, SQE, LL, TU, JAB, EXC, CRS */
    write8(ioaddr + XIRCREG40_TMASK1, 0xb0); /* rsv, rsv, PTD, EXT, rsv,rsv,rsv, rsv*/
    write8(ioaddr + XIRCREG40_RXST0,  0x00); /* ROK, RAB, REN, RO, CRC, AE, PTL, MP */
    write8(ioaddr + XIRCREG40_TXST0,  0x00); /* TOK, TAB, SQE, LL, TU, JAB, EXC, CRS */
    write8(ioaddr + XIRCREG40_TXST1,  0x00); /* TEN, rsv, PTD, EXT, retry_counter:4  */
		
	if (local->mohawk && init_mii(device)) {
		if (device->media == 4 || local->new_mii || local->dingo) {
	    	SelectPage(2);
	    	write8(ioaddr + XIRCREG2_MSR, read8(ioaddr + XIRCREG2_MSR) | 0x08);
	    	snooze(HZ/50);
		} else {
			/* MII: 10 Mbps */
	    	SelectPage(0x42);
	    	if (device->media == 2) /* enable 10Base2 */
				write8(ioaddr + XIRCREG42_SWC1, 0xC0);
	    	else  /* enable 10BaseT */
				write8(ioaddr + XIRCREG42_SWC1, 0x80);
	    	snooze(HZ/25);	/* wait 40 msec to let it complete */
	    }
    } else {  /* No MII*/
	    unsigned value;
		SelectPage(0);
		value = read8(XIRCREG_ESR);	 /* read the ESR */
		device->media = (value & MediaSelect) ? 1 : 2;
    }

    /* configure the LEDs */
    SelectPage(2);
    if (device->media == 1 || device->media == 4) /* TP: Link and Activity */
		write8(ioaddr + XIRCREG2_LED, 0x3b);
    else
		write8(ioaddr + XIRCREG2_LED, 0x3a);

	if (local->dingo)
		write8(ioaddr + 0x0b, 0x04); /* 100 Mbit LED */

	/* Enable MAC */
	SelectPage(0x40);
	write8(ioaddr + XIRCREG40_CMD0, EnableRecv | Online);
	
	/* setup Ethernet IMR and enable interrupts */
    SelectPage(1);
    write8(ioaddr + XIRCREG1_IMR0, 0xff); /* All interrupts */
    snooze(1); /* Wait for a msec */
    SelectPage(0);
    write8(ioaddr + XIRCREG_CR, EnableIntr); /* Enable interrupts */
	if (local->modem && !local->dingo) { /* do some magic */
		if (!(read8(ioaddr + 0x10) & 0x01))
		    write8(ioaddr + 0x10, 0x11); /* unmask master-int bit */
    }
	
    SelectPage(0);
	ETHER_DEBUG(device, FUNCTION,kDevName ": reset_device reset the NIC hardware\n"); 
};

static void dump_packet( const char * msg, unsigned char * buf, uint16 size) {

	uint16 j;
	
	dprintf("%s dumping %x size %d \n", msg, buf, size);
	for (j=0; j<size; j++) {
		if ((j & 0xF) == 0) dprintf("\n");
		dprintf("%2.2x ", buf[j]);
	}
}

static void dump_rx_desc(ether_dev_info_t *device) {
#if 0
	uint16 j;
	/* kprintf is used here since this is called from the serial debug command */
	kprintf("rx desc %8.8x \n", device->rx_desc);
	for (j=0; j < RX_BUFFERS; j++ ) {
		kprintf("rx_desc[%2.2d]=...\n", j);
	}
#endif
}

static void dump_tx_desc(ether_dev_info_t *device) {
}

uint8 load_irq_settings(uint16 *irq, uint32 *port) {

	void *handle;
	const driver_settings *settings;
	uint8 found = 0;
	
	handle = load_driver_settings(kDevName);
	settings = get_driver_settings(handle);
	if (settings) {
		int i;
		for (i=0;i<settings->parameter_count;i++) {
			driver_parameter *parameter = settings->parameters + i;
			if (strcmp(parameter->name, "irq") == 0) {
				*irq = strtoll(parameter->values[0], NULL, 0);
				dprintf(kDevName " forcing irq to 0x%x\n", *irq);
				found |= 0x1;
			} else if( strcmp(parameter->name, "port") == 0) {			
				*port = strtoll(parameter->values[0], NULL, 0);
				dprintf(kDevName " forcing port to 0x%x\n", *port);
				found |= 0x2;
			}
		}
	}
	unload_driver_settings(handle);
	if (( found == 1) || (found == 2)) dprintf (kDevName ": must specify port AND irq\n");
	if (found == 0x3) return true; else return false;
}


/* Serial Debugger command
   Connect a terminal emulator to the serial port at 19.2 8-1-None
   Press the keys ( alt-sysreq on Intel) or (Clover-leaf Power on Mac ) to enter the debugger
   At the kdebug> prompt enter "tcm5x9 arg...",
   for example "AcmeEthernet R" to enable a received packet trace.
*/
#if DEBUGGER_COMMAND
static int
xircom(int argc, char **argv) {
	uint16 i,j;
	const char * usage = "usage: xircom { Function_calls | Interrupts | Number_frames | BUS_IO | Stats | Rx_trace | Tx_trace }\n";	
	

	if (argc < 2) {
		kprintf("%s",usage);	return 0;
	}

	for (i= argc, j= 1; i > 1; i--, j++) {
		switch (*argv[j]) {
		case 'F':
		case 'f':
			gdev->debug ^= FUNCTION;
			if (gdev->debug & FUNCTION) 
				kprintf("Function() call trace Enabled\n");
			else 			
				kprintf("Function() call trace Disabled\n");

		case 'N':
		case 'n':
			gdev->debug ^= SEQ;
			if (gdev->debug & SEQ) 
				kprintf("Sequence numbers packet trace Enabled\n");
			else 			
				kprintf("Sequence numbers packet trace Disabled\n");
			break; 
		case 'R':
		case 'r':
			gdev->debug ^= RX;
			if (gdev->debug & RX) 
				kprintf("Receive packet trace Enabled\n");
			else 			
				kprintf("Receive packet trace Disabled\n");
			break;
		case 'T':
		case 't':
			gdev->debug ^= TX;
			if (gdev->debug & TX) 
				kprintf("Transmit packet trace Enabled\n");
			else 			
				kprintf("Transmit packet trace Disabled\n");
			break; 
		case 'S':
		case 's':
			kprintf(kDevName " statistics\n");			
//			kprintf("received %d,  transmitted %d\n", gdev->stats_rx, gdev->stats_tx);
//			kprintf("rx err %d,  tx err %d\n", gdev->stats_rx_err, gdev->stats_tx_err);
//			kprintf("rx overrun %d,  tx underrun %d\n", gdev->stats_rx_overrun, gdev->stats_tx_underrun);
//			kprintf("collisions %d\n", gdev->stats_collisions);
//			kprintf("handled %d and unhandled %d interrupts\n", gdev->handled_int, gdev->unhandled_int);
			break; 
		case 'b':
		case 'B':
			gdev->debug ^= BUS_IO;
			if (gdev->debug & BUS_IO) 
				kprintf("PCI IO trace Enabled\n");
			else 			
				kprintf("PCI IO trace Disabled\n");
			break; 
		default:
			kprintf("%s",usage);
			return 0;
		}
	}
	
	return 0;
}
#endif /* DEBUGGER_COMMAND */

static void clear_statistics(ether_dev_info_t *device) {
	
	device->handled_int = device->unhandled_int = 0;
	device->stats_rx = device->stats_tx = 0;
//	device->stats_rx_err = device->stats_tx_err = 0;
	device->stats_rx_overrun = device->stats_tx_underrun = 0;
	device->stats_collisions = 0;
}

/*
 * Allocate and initialize semaphores and spinlocks.
 */
static status_t allocate_resources(ether_dev_info_t *device) {

	/* Setup Semaphores */
	if ((device->ilock = create_sem(0, kDevName " rx")) < 0) {
		dprintf(kDevName " create rx sem failed %x \n", device->ilock);
		return (device->ilock);
	}
	set_sem_owner(device->ilock, B_SYSTEM_TEAM);
	
	if ((device->olock = create_sem(1, kDevName " tx")) < 0) {
		delete_sem(device->ilock);
		dprintf(kDevName " create tx sem failed %x \n", device->olock);
		return (device->olock);
	}
	set_sem_owner(device->olock, B_SYSTEM_TEAM);

	device->readLock = device->writeLock = 0;

	device->blockFlag = 0;
	return (B_OK);
}

static void free_resources(ether_dev_info_t *device) {
		delete_sem(device->ilock);
		delete_sem(device->olock);
}



/*** PCMCIA STUFF ***/

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/
#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
static const char *version = "xircom.c 0.04b";
#define DEBUG(n, args) do { if (pc_debug>(n)) printk args; } while (0)
#else
#define DEBUG(n, args) do { } while (0)
#endif

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* List of interrupts to choose from */
static int irq_list[4] = { -1 };

/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card
   insertion and ejection events.  They are invoked from the dummy
   event handler. 
*/

static void dummy_config(dev_link_t *link);
static void dummy_release(u_long arg);
static int dummy_event(event_t event, int priority,
		       event_callback_args_t *args);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *dummy_attach(void);
static void dummy_detach(dev_link_t *);

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "xircom";

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally shouldn't be allocated dynamically.

   In this case, we also provide a flag to indicate if a device is
   "stopped" due to a power management event, or card ejection.  The
   device IO routines can use a flag like this to throttle IO to a
   card that is not ready to accept it.
*/
   
/*
  BeOS bus manager declarations: we use the shorthand references to
  bus manager functions to make the code somewhat more portable across
  platforms.
*/
static cs_client_module_info *cs = NULL;
static ds_module_info *ds = NULL;
#define CardServices		cs->_CardServices
#define register_pccard_driver	ds->_register_pccard_driver
#define unregister_pccard_driver ds->_unregister_pccard_driver

#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define next_tuple(a, b, c)  get_tuple(GetNextTuple, a, b, c)

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err;
    err.func = func; err.retcode = ret;
    CardServices(ReportError, handle, &err);
}

static int
get_tuple_data(int fn, client_handle_t handle, tuple_t *tuple)
{
    int err;

    if ((err=CardServices(fn, handle, tuple)))
	return err;
    return CardServices(GetTupleData, handle, tuple);
}

static int
get_tuple(int fn, client_handle_t handle, tuple_t *tuple, cisparse_t *parse)
{
    int err;

    if ((err=get_tuple_data(fn, handle, tuple)))
	return err;
    return CardServices(ParseTuple, handle, tuple, parse);
}



/*======================================================================

    dummy_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/

static dev_link_t *dummy_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    local_info_t *local;
    int ret, i, j;
    DEBUG(0, ("dummy_attach()\n"));

    /* Find a free slot in the device table */
    for (i = 0; i < MAX_DEVS; i++)
	if (devs[i] == NULL) break;
    if (i == MAX_DEVS) {
	printk("dummy_cs: no devices available\n");
	return NULL;
    }
	
	/* Initialize the dev_link_t structure */
    link = malloc(sizeof(struct dev_link_t));
	if (link == NULL) {
		return NULL;
	}
	memset(link, 0, sizeof(struct dev_link_t));	
    devs[i] = link;
    ndevs++;
    link->release.function = &dummy_release;
    link->release.data = (u_long)link;
	
    /* Interrupt setup */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;
    link->irq.Handler = NULL; /* Done by install_io_interrupt */    
    /*
      General socket configuration defaults can go here.  In this
      client, we assume very little, and rely on the CIS for almost
      everything.  In most clients, many details (i.e., number, sizes,
      and attributes of IO windows) are fixed by the nature of the
      device, and can be hard-wired here.
    */

    /* Allocate space for private device-specific data */
    local = malloc(sizeof(local_info_t));
    memset(local, 0, sizeof(local_info_t));
    link->priv = local;
	/* Register with Card Services */
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &dummy_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != 0) {
	cs_error(link->handle, RegisterClient, ret);
	dummy_detach(link);
	return NULL;
    }

    return link;
} /* dummy_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void dummy_detach(dev_link_t *link)
{
    int i;
    
    DEBUG(0, ("dummy_detach(0x%p)\n", link));
    
	if(link == NULL) {
		dprintf("dummy_detach called with NULL argumant\n");
		return;
	}

    /* Locate device structure */
    for (i = 0; i < MAX_DEVS; i++)
	if (devs[i] == link) break;
    if (i == MAX_DEVS)
	return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
	DEBUG(0, ("dummy_cs: detach postponed, '%s' "
		  "still locked\n", "xircom"));
	link->state |= DEV_STALE_LINK;
	return;
    }

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    devs[i] = NULL;
    if (link->priv)
	free(link->priv);
    free(link);
    ndevs--;
    
} /* dummy_detach */

static int
has_ce2_string(dev_link_t * link)
{
    client_handle_t handle = link->handle;
    tuple_t tuple;
    cisparse_t parse;
    u_char buf[256];

    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = 254;
    tuple.TupleOffset = 0;
    tuple.DesiredTuple = CISTPL_VERS_1;
    if (!first_tuple(handle, &tuple, &parse) && parse.version_1.ns > 2) {
	if (strstr(parse.version_1.str + parse.version_1.ofs[2], "CE2"))
	    return 1;
    }
    return 0;
}


/****************
 * Detect the type of the card. s is the buffer with the data of tuple 0x20
 * Returns: 0 := not supported
 *		       mediaid=11 and prodid=47
 * Media-Id bits:
 *  Ethernet	    0x01
 *  Tokenring	    0x02
 *  Arcnet	    0x04
 *  Wireless	    0x08
 *  Modem	    0x10
 *  GSM only	    0x20
 * Prod-Id bits:
 *  Pocket	    0x10
 *  External	    0x20
 *  Creditcard	    0x40
 *  Cardbus	    0x80
 *
 */
static int
set_card_type(dev_link_t *link, const void *s)
{
    local_info_t *local = link->priv;
    unsigned mediaid= ((const unsigned char *)s)[3];
    unsigned prodid = ((const unsigned char *)s)[4];

     local->mohawk = 0;
     local->dingo = 0;
     local->modem = 0;
 
    local->card_type = XIR_UNKNOWN;
    if (!(prodid & 0x40)) {
	dprintf("xircom: set_card_type:: Not a creditcard\n");
	return 0;
    }
    if (!(mediaid & 0x01)) {
	dprintf("xircom: set_card_type:: Not an Ethernet card\n");
	return 0;
    }
    if (mediaid & 0x10) {
	local->modem = 1;
	switch(prodid & 15) {
	  case 1: local->card_type = XIR_CEM   ; break;
	  case 2: local->card_type = XIR_CEM2  ; break;
	  case 3: local->card_type = XIR_CEM3  ; break;
	  case 4: local->card_type = XIR_CEM33 ; break;
	  case 5: local->card_type = XIR_CEM56M;
		local->mohawk = 1;
		  break;
	  case 6:
	  case 7: /* 7 is the RealPort 10/56 */
		  dprintf("xircom: set_card_type:: CEM56 found: experimental support\n");
		  local->card_type = XIR_CEM56;
		  local->mohawk = 1; 
		  local->dingo = 1;
		  break;
	}
    } else {
	switch(prodid & 15) {
	  case 1: local->card_type = has_ce2_string(link)? XIR_CE2 : XIR_CE ;
		  break;
	  
	  case 2: local->card_type = XIR_CE2; break;
	  case 3: local->card_type = XIR_CE3;
		  local->mohawk = 1;
		  dprintf("xircom: set_card_type:: CE3 found\n");
		  break;
	}
    }
    if (local->card_type == XIR_CE || local->card_type == XIR_CEM) {
		dprintf("xircom: set_card_type:: Sorry, this is an old CE card\n");
	return 0;
    }
    if (local->card_type == XIR_UNKNOWN)
		dprintf("xircom: set_card_type:: unknown card (mediaid=%02x prodid=%02x)\n", mediaid, prodid);
    return 1;
}


/*======================================================================

    dummy_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.
    
======================================================================*/

#define CS_CHECK(fn, h, a) \
while ((last_ret=CardServices(last_fn=(fn),h,a))!=0) goto cs_failed
#define CS_CHECK2(fn, h, a, b) \
while ((last_ret=CardServices(last_fn=(fn),h,a,b))!=0) goto cs_failed

#define CFG_CHECK(fn, h, a) \
if (CardServices(fn, h, a) != 0) goto next_entry
#define CFG_CHECK2(fn, h, a, b) \
if (CardServices(fn, h, a, b) != 0) goto next_entry

static void dummy_config(dev_link_t *link)
{
	client_handle_t handle;
	tuple_t tuple;
	cisparse_t parse;
	local_info_t *dev;
	int last_fn, last_ret, nd;
	u_char buf[64];
	win_req_t req;
	memreq_t map;
	ioaddr_t ioaddr;
	int i;
    cistpl_lan_node_id_t *node_id = (cistpl_lan_node_id_t*)parse.funce.data;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;	
	int err;
    
	
	handle = link->handle;
	dev = link->priv;
	dev->dingo_ccr = 0;
	
	DEBUG(0, ("dummy_config(0x%p)\n", link));
	 /*
     * This reads the card's CONFIG tuple to find its configuration
     * registers.
     */
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = 64;
    tuple.TupleOffset = 0;

    /* Is this a valid	card */
    tuple.DesiredTuple = CISTPL_MANFID;
    if ((err=first_tuple(handle, &tuple, &parse))) {
	goto failure;
    }

    if (!(parse.manfid.manf == MANFID_XIRCOM)) { 
		DEBUG(0, "xircom: no xircom card found\n");
		goto failure;
	}

    if (!set_card_type(link, buf)) {
		dprintf("xircom: Sorry, this card is not supported\n");
		goto failure;
    }

	/* get configuration stuff */
    tuple.DesiredTuple = CISTPL_CONFIG;
    if ((err=first_tuple(handle, &tuple, &parse)))
	goto cis_error;
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present =    parse.config.rmask[0];

    /* get the hardware address from the CIS */
    tuple.DesiredTuple = CISTPL_FUNCE;
    for (err = first_tuple(handle, &tuple, &parse); !err;
			     err = next_tuple(handle, &tuple, &parse)) {
	/* Once I saw two CISTPL_FUNCE_LAN_NODE_ID entries:
	 * the first one with a length of zero the second correct -
	 * so I skip all entries with length 0 */
	if (parse.funce.type == CISTPL_FUNCE_LAN_NODE_ID
	    && ((cistpl_lan_node_id_t *)parse.funce.data)->nb)
	    break;
    }

    if (err) { /* not found: try to get the node-id from tuple 0x89 */
	tuple.DesiredTuple = 0x89;  /* data layout looks like tuple 0x22 */
	if (!(err = get_tuple_data(GetFirstTuple, handle, &tuple))) {
	    if (tuple.TupleDataLen == 8 && *buf == CISTPL_FUNCE_LAN_NODE_ID)
		memcpy(&parse, buf, 8);
	    else
		err = -1;
	}
    }
    if (err) {
		dprintf("xircom: node-id not found in CIS\n");
		goto failure;
    }
    node_id = (cistpl_lan_node_id_t *)parse.funce.data;
    if (node_id->nb != 6) {
		dprintf("xircom: malformed node-id in CIS\n");
		goto failure;
    }
    /* Set MAC-addr here (any other way?) */
    for (i=0; i < 6; i++) {
		dev->macAddr[i] = node_id->id[i];
	}
	
	/* Configure card */
    link->state |= DEV_CONFIG;

    link->io.IOAddrLines =10;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
    link->irq.Attributes = IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
    if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
    else {
	for (i = 0; i < 4; i++)
	    link->irq.IRQInfo2 |= 1 << irq_list[i];
    }
	if (dev->modem) {
		int pass;

		link->irq.Attributes |= IRQ_TYPE_DYNAMIC_SHARING|IRQ_FIRST_SHARED ;
		link->io.NumPorts2 = 8;
		link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
		if (dev->dingo) {
	    	/* Take the Modem IO port from the CIS and scan for a free
	     	* Ethernet port */
	    	link->io.NumPorts1 = 16; /* no Mako stuff anymore */
	    	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	    	for (err = first_tuple(handle, &tuple, &parse); !err; err = next_tuple(handle, &tuple, &parse)) {
					if (cf->io.nwin > 0  &&  (cf->io.win[0].base & 0xf) == 8) {
			    		for (ioaddr = 0x300; ioaddr < 0x400; ioaddr += 0x10) {
							link->conf.ConfigIndex = cf->index ;
							link->io.BasePort2 = cf->io.win[0].base;
							link->io.BasePort1 = ioaddr;
							if (!(err=CardServices(RequestIO, link->handle, &link->io)))
			    				goto port_found;
		    			}
					}
	    	}
		} else /* No dingo */{
	    link->io.NumPorts1 = 18;
	    /* We do 2 passes here: The first one uses the regular mapping and
	     * the second tries again, thereby considering that the 32 ports are
	     * mirrored every 32 bytes. Actually we use a mirrored port for
	     * the Mako if (on the first pass) the COR bit 5 is set.
	     */
	    for (pass=0; pass < 2; pass++) {
			tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
			for (err = first_tuple(handle, &tuple, &parse); !err;
			     err = next_tuple(handle, &tuple, &parse)){
		    		if (cf->io.nwin > 0  &&  (cf->io.win[0].base & 0xf) == 8){
					link->conf.ConfigIndex = cf->index ;
					link->io.BasePort2 = cf->io.win[0].base;
					link->io.BasePort1 = link->io.BasePort2 + (pass ? (cf->index & 0x20 ? -24:8) : (cf->index & 0x20 ?   8:-24));
					if (!(err=CardServices(RequestIO, link->handle,&link->io)))
			    		goto port_found;
		   			}
			}
	    }
	    /* if special option:
	     * try to configure as Ethernet only.
	     * .... */
	}
	dprintf("No ports available\n");
	} else /*No modem */ {
	
		link->irq.Attributes |= IRQ_TYPE_EXCLUSIVE;
		link->io.NumPorts1 = 16;
	
		/* Probe for Xircom card in range 0x300-0x400 */
		for (ioaddr = 0x300; ioaddr < 0x400; ioaddr += 0x10) {
	    	link->io.BasePort1 = ioaddr;
	    	if (!(err=CardServices(RequestIO, link->handle, &link->io)))
			goto port_found;
		}
		/* Maybe CS can tell us */
		
		link->io.BasePort1 = 0; /* let CS decide */
		if ((err=CardServices(RequestIO, link->handle, &link->io))) {
	    	cs_error(link->handle, RequestIO, err);
	   		goto config_error;
		}
	}
	
  port_found:
    if (err)
	 goto config_error;

    /****************
     * Now allocate an interrupt line.	Note that this does not
     * actually assign a handler to the interrupt.
     */
    if ((err=CardServices(RequestIRQ, link->handle, &link->irq))) {
		cs_error(link->handle, RequestIRQ, err);
		goto config_error;
    }

    /****************
     * This actually configures the PCMCIA socket -- setting up
     * the I/O windows and the interrupt mapping.
     */
    if ((err=CardServices(RequestConfiguration,
			  link->handle, &link->conf))) {
		cs_error(link->handle, RequestConfiguration, err);
		goto config_error;
    }
    if (dev->dingo) {
		conf_reg_t reg;
		win_req_t req;
		memreq_t mem;

		/* Reset the modem's BAR to the correct value
		 * This is necessary because in the RequestConfiguration call,
	 	* the base address of the ethernet port (BasePort1) is written
	 	* to the BAR registers of the modem.
	 	*/
		reg.Action = CS_WRITE;
		reg.Offset = CISREG_IOBASE_0;
		reg.Value = link->io.BasePort2 & 0xff;
		if ((err = CardServices(AccessConfigurationRegister, link->handle, &reg))) {
	   	 cs_error(link->handle, AccessConfigurationRegister, err);
	   	 goto config_error;
		}
		reg.Action = CS_WRITE;
		reg.Offset = CISREG_IOBASE_1;
		reg.Value = (link->io.BasePort2 >> 8) & 0xff;
		if ((err = CardServices(AccessConfigurationRegister, link->handle, &reg))) {
	    	cs_error(link->handle, AccessConfigurationRegister, err);
	    	goto config_error;
		}
     
		/* There is no config entry for the Ethernet part which
	 	* is at 0x0800. So we allocate a window into the attribute
	 	* memory and write direct to the CIS registers
	 	*/
		req.Attributes = WIN_DATA_WIDTH_8|WIN_MEMORY_TYPE_AM|WIN_ENABLE;
		req.Base = 0;
		req.Size = 0x1000; /* 4k window */
		req.AccessSpeed = 0;
		link->win = (window_handle_t)link->handle;
		if ((err = CardServices(RequestWindow, &link->win, &req))) {
	    	cs_error(link->handle, RequestWindow, err);
	    	goto config_error;
		}
		dev->dingo_ccr = ioremap(req.Base,0x1000) + 0x0800;
		mem.CardOffset = 0x0;
		mem.Page = 0;
		if ((err = CardServices(MapMemPage, link->win, &mem))) {
	    	cs_error(link->handle, MapMemPage, err);
	    	goto config_error;
		}

		/* Setup the CCRs; there are no infos in the CIS about the Ethernet
		 * part.
		 */
		writeb(0x47, dev->dingo_ccr + CISREG_COR);
		ioaddr = link->io.BasePort1;
		writeb(ioaddr & 0xff	  , dev->dingo_ccr + CISREG_IOBASE_0);
		writeb((ioaddr >> 8)&0xff , dev->dingo_ccr + CISREG_IOBASE_1);


		writeb(0x01, dev->dingo_ccr + 0x20);
		writeb(0x0c, dev->dingo_ccr + 0x22);
		writeb(0x00, dev->dingo_ccr + 0x24);
		writeb(0x00, dev->dingo_ccr + 0x26);
		writeb(0x00, dev->dingo_ccr + 0x28);
}

	for (nd = 0; nd < MAX_DEVS; nd++)
		if (devs[nd] == link) break;
		
	sprintf(dev->node.dev_name, "net/" kDevName "/%d", nd);
	dev->node.major = dev->node.minor = 0;
	link->dev = &dev->node;
	
    link->state &= ~DEV_CONFIG_PENDING;
	/* FIXME: Maybe we should reset the cem56 here as in the linux driver*/
	/* Is it normal for a card to need two resets to work?! */
	return;
	
  config_error:
    dprintf("xircom: config_error in dummy_config\n");
    link->state &= ~DEV_CONFIG_PENDING;
    dummy_release((u_long)link);
    return;

  cis_error:
    printk("xircom: unable to parse CIS\n");
  failure:
    link->state &= ~DEV_CONFIG_PENDING;

} /* dummy_config */

/*======================================================================

    After a card is removed, dummy_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void dummy_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;

    DEBUG(0, ("dummy_release(0x%p)\n", link));

    /*
       If the device is currently in use, we won't release until it
       is actually closed, because until then, we can't be sure that
       no one will try to access the device or its data structures.
    */
    if (link->open) {
	DEBUG(1, ("dummy_cs: release postponed, '%s' still open\n",
		  link->dev->dev_name));
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    /* Unlink the device chain */
    link->dev = NULL;

    /*
      In a normal driver, additional code may be needed to release
      other kernel data structures associated with this device. 
    */
    
    /* Don't bother checking to see if these succeed or not */
    if (link->win) {
		local_info_t *local = link->priv;
		if (local->dingo)
		    iounmap(local->dingo_ccr - 0x0800);
			CardServices(ReleaseWindow, link->win);
    }
    
    CardServices(ReleaseConfiguration, link->handle);
    
    if (link->io.NumPorts1)
		CardServices(ReleaseIO, link->handle, &link->io);
    
    if (link->irq.AssignedIRQ)
		CardServices(ReleaseIRQ, link->handle, &link->irq);
    
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
		dummy_detach(link);
    
} /* dummy_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.

    When a CARD_REMOVAL event is received, we immediately set a
    private flag to block future accesses to this device.  All the
    functions that actually access the device should check this flag
    to make sure the card is still present.
    
======================================================================*/

static int dummy_event(event_t event, int priority,
		       event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;

    DEBUG(1, ("dummy_event(0x%06x)\n", event));
    
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
		link->state &= ~DEV_PRESENT;
		if (link->state & DEV_CONFIG) {
	    	((local_info_t *)link->priv)->stop = 1;
	    	link->release.expires = RUN_AT(HZ/20);
	    	cs->_add_timer(&link->release);
		}
		break;
    case CS_EVENT_CARD_INSERTION:
		link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
		dummy_config(link);
		break;
    case CS_EVENT_PM_SUSPEND:
		link->state |= DEV_SUSPEND;
		/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
		/* Mark the device as stopped, to block IO until later */
		((local_info_t *)link->priv)->stop = 1;
		if (link->state & DEV_CONFIG)
	    	CardServices(ReleaseConfiguration, link->handle);
		break;
    case CS_EVENT_PM_RESUME:
		link->state &= ~DEV_SUSPEND;
		/* Fall through... */
    case CS_EVENT_CARD_RESET:
		if (link->state & DEV_CONFIG)
	    	CardServices(RequestConfiguration, link->handle, &link->conf);
		((local_info_t *)link->priv)->stop = 0;
	/*
	  In a normal driver, additional code may go here to restore
	  the device state and restart IO. 
	*/
		break;
    }
    return 0;
} /* dummy_event */


status_t pc_init_driver(void)
{
    client_reg_t client_reg;
    servinfo_t serv;
    bind_req_t bind;
    int i, ret;
    
    for(i=0;i<MAX_DEVS;i++) devs[i] = NULL;
    
    if (get_module(CS_CLIENT_MODULE_NAME, (struct module_info **)&cs) != B_OK) goto err1;
    if (get_module(DS_MODULE_NAME, (struct module_info **)&ds) != B_OK) goto err2;
	
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) goto err3;

    /* When this is called, Driver Services will "attach" this driver
       to any already-present cards that are bound appropriately. */
	register_pccard_driver(&dev_info, &dummy_attach, &dummy_detach);
    return B_OK;
err3:	
	put_module(DS_MODULE_NAME);
err2:
	put_module(CS_CLIENT_MODULE_NAME);
err1:
	return B_ERROR;
}


void pc_uninit_driver(void)
{
    int i;
    if(cs && ds){
 		unregister_pccard_driver(&dev_info);
        for (i = 0; i < MAX_DEVS; i++)
        if (devs[i]) {
            if (devs[i]->state & DEV_CONFIG)
            dummy_release((u_long)devs[i]);
            dummy_detach(devs[i]);
        }
        if (ds) put_module(DS_MODULE_NAME);
        if (cs) put_module(CS_CLIENT_MODULE_NAME);
    }
}


const char **publish_devices(void)
{
    int i, nd;

    for (nd = i = 0; nd < MAX_DEVS; nd++) {
		if (devs[nd] == NULL) continue;
		gDevNameList[i] = (char *)malloc(strlen(devs[nd]->dev->dev_name)+1);
		strcpy(gDevNameList[i++], devs[nd]->dev->dev_name);
    }
    gDevNameList[i] = NULL;
    return (const char **)gDevNameList;
}

int32	api_version = B_CUR_DRIVER_API_VERSION;
